/********************************************************
 * backup -- Backup a bunch of files to a device	*
 ********************************************************/
#include <getopt.h>
#include <stdlib.h>
#include <fstream>
#include <sys/param.h>
#include <unistd.h>

#include "tty_lock.h"
#include "config.h"
#include "tmp_file.h"
#include "filer.h"
#include "input_list.h"
#include "io_b.h"
#include "batch_b.h"
#include "log.h"
#include "exit.h"
#include "util.h"
#include "backup.h"

config global_config;		// The configuration
meta_b global_meta;		// The global configuration information

bool keep_option = false;	// Keep temp files
bool image_option = false;	// Keep image files

bool do_log_option = false;	// If set, log things


std::ofstream log_file;		// File for logging things
std::string backup_root("/");	// Root of the backups

unsigned int block_restriction = 0;	// Blocking restriction for the system


// Directory from which we started this mess
static char old_dir[MAXPATHLEN];

typedef std::vector<io_b*> io_array;	// Set of I/O functions

bool tty_lock::locked = false;	// TTY is currently locked
/********************************************************
 * usage -- Tell the user what to do			*
 ********************************************************/
static void usage(void)
{
    std::cerr << "Usage is backup [options] <config-file> <job> [<job> ...]\n";
    std::cerr << "Options:\n";
    std::cerr << "	-k -- Keep temporary files\n";
    std::cerr << "	-i -- Keep image files\n";
    std::cerr << "	-v -- Verbose output\n";
    std::cerr << "	-l<log-file> -- Specify file for logging things\n";
    std::cerr << "	-d<str> -- Set log file debugging level\n";
    std::cerr << "	-D<str> -- Set screen debugging level\n";
    exit (EXIT_FATAL);
}
/********************************************************
 * do_backup -- Do a single backup			*
 ********************************************************/
static void do_backup(
	const std::string& name	// Name of the backup
) {
    std::cout << "Backup: " << name << std::endl;

    // The information about the backup
    single_config backup = global_config.get_single(SECTION_BACKUP, name);

    filer the_filer;	// The file list handing function

    // The list of all input_list directive
    string_vector input_list_list = backup.get_option_all(SECTION_INPUT_LIST);

    // Process each of the input items
    for (string_vector::const_iterator input_name = input_list_list.begin();
	 input_name != input_list_list.end();
	 ++input_name) {

	// The inputlist module for this InputList directive
	input_list list(*input_name, global_config);

	// The files from the input list
	string_set the_files = list.find_files();
	the_filer.add_files(the_files);
    }

    // The list of batchers to use
    string_vector batch_list = backup.get_option_all(SECTION_BATCH);

    // The list of I/O systems to use
    io_array io_list;

    // The names of each of the I/O systems
    string_vector io_names = backup.get_option_all(SECTION_IO);
    if (io_names.size() == 0) {
	USER_ERROR("No IO Section specified");
    }

    // Loop through and create each I/O system
    for (string_vector::const_iterator cur_io = io_names.begin();
	    cur_io != io_names.end();
	    ++cur_io) {

	// Create the I/O record
	io_b* new_io = new_io_b(*cur_io, global_config);

	// Check to see if we have any additional blocking restrictions
	if (new_io->block_restriction() != 0) {
	    if (new_io->block_restriction() > block_restriction) {
		if ((new_io->block_restriction() % block_restriction) != 0) {
		    USER_ERROR(
			"Incompatible blocking restrictions for the I/O systems");
		}
		block_restriction = new_io->block_restriction();
	    } else {
		if ((block_restriction % new_io->block_restriction()) != 0) {
		    USER_ERROR(
			"Incompatible blocking restrictions for the I/O systems");
		}
	    }
	}

	io_list.push_back(new_io);
    }

    /*
     * Loop through each batch type and try and stuff files into
     * the batch object (and get batches as a result)
     */
    for (string_vector::const_iterator batch_name = batch_list.begin();
	    batch_name != batch_list.end();
	    ++batch_name) {
	// The batch that we are using now
	batch_b* batch = new_batch(*batch_name, global_config);

	while (1) { 
	    // Add the files to the batch and then check results
	    batch_b::status status = batch->add_files(the_filer);

	    if (status == batch_b::BATCH_NONE)
		break;

	    while (1) {
		// Create a batch and return information on it
		batch_info_b* info = batch->create_batch();

		if (info == NULL)
		    break;

		//TODO2: Handle multiple I/O systems
		io_list[0]->write(info);
	    }
	}
	delete batch;
	batch = NULL;
    }
    // Name of the meta file
    std::string meta_file = tmp_name("meta");

    // Name of the file to put the meta eport in
    std::string meta_report = tmp_name("meta_report");
    
    // Finish all the I/O stuff
    for (io_array::const_iterator cur_io = io_list.begin();
	    cur_io != io_list.end();
	    ++cur_io) {
	(*cur_io)->finish(meta_file, meta_report);
    }
    if (! keep_option) {
	if (unlink(meta_report.c_str()) != 0)
	    SYSTEM_ERROR("Unable to unlink " + meta_report);
	if (unlink(meta_file.c_str()) != 0)
	    SYSTEM_ERROR("Unable to unlink " + meta_file);
    }
}

/********************************************************
 * do_job -- Process a single backup job		*
 ********************************************************/
static void do_job(
    const char *const job_name
) {
    std::cout << "Job: " << job_name << std::endl;

    // The information about this job
    single_config job = global_config.get_single(SECTION_JOB_LIST, job_name);

    // This list of jobs we need to do
    string_vector job_list = job.get_list(JOB_LIST_JOB);

    // Loop through each backup job and do it
    for (string_vector::iterator cur_job = job_list.begin();
	    cur_job != job_list.end();
	    ++cur_job) {
	do_backup(*cur_job);
    }
}

/*------------------------------------------------------*/
int main(int argc, char *argv[])
{
    try {
	// Name of the log file (if any)
	char *log_file_name = NULL;

	while (1) {
	    int opt = getopt(argc, argv, "d:D:kil:");
	    if (opt == EOF)
		break;

	    switch (opt) {
		case 'd':
		    init_log(optarg);
		    break;
		case 'D':
		    init_console_log(optarg);
		    break;
		case 'k':
		    keep_option = true;
		    break;
		case 'i':
		    image_option = true;
		    break;
		case 'l':
		    log_file_name = optarg;
		    do_log_option = true;
		    break;
		default:
		    std::cerr << "Illegal option " << opt << std::endl;
		    usage();
		    break;
	    }
	}

	if ((argc - optind) < 2) {
	    usage();
	}
	if (log_file_name != NULL) {
	    log_file.open(log_file_name, std::ios::out);
	    if (log_file.bad()) {
		std::cerr << "Error: Could not open log file " << 
		    log_file_name << std::endl;
		exit (EXIT_FATAL);
	    }
	}

	global_config.read(argv[optind]);
	tmp_name_init(global_config);

	if (global_config.defined(SECTION_OPTIONS, OPTION_GLOBAL, 
		    OPTION_BACKUP_ROOT, 0)) {

	    backup_root = global_config.get_option(
		    SECTION_OPTIONS, OPTION_GLOBAL, OPTION_BACKUP_ROOT);
	}
	
	if (getcwd(old_dir, sizeof(old_dir)) == NULL) {
	    SYSTEM_ERROR("getcwd failed");
	}

	// Consume an argument
	++optind;
	while (optind < argc) {
	    do_job(argv[optind]);
	    ++optind;
	}
	if (chdir(old_dir) != 0) {
	    SYSTEM_ERROR("chdir failed");
	}
    } 
    catch (user_error &err) {
	err.print();
    }
    catch (internal_error &err) {
	err.print();
    }
    catch (system_error &err) {
	err.print();
    }
    catch (generic_error &err) {
	std::cerr << "Internal error: Strange error caught" << std::endl;
	err.print();
    }
    return (EXIT_SUCCESS);
}


