#ifndef __EXCEPTION_H__
#define __EXCEPTION_H__
/********************************************************
 * There are three kinds of exceptions:			*
 * 	user_error -- User did something wrong		*
 * 	system_error -- System call failed		*
 * 	internal_error -- Something impossible happened	*
 ********************************************************/

#include <string>
#include <iostream>
#include <errno.h>

/********************************************************
 * generic_error -- General type of error		*
 ********************************************************/
class generic_error {
    public:
	const char *const file;	// File in which the error occurred
	const int line;		// The line where the error occurred
	const std::string type; // Type of error (user, system, internal)
	const std::string error;// Error message

	explicit generic_error(
	    const char *const i_file,
	    const int i_line,
	    const std::string &i_type,
	    const std::string &i_error
	):
	    file(i_file),
	    line(i_line),
	    type(i_type),
	    error(i_error)
	{};
	// Copy constructor defaults
	// Assignment operator defaults
	// Destructor defaults
	void print(void) {
	    std::cerr << type << " error at " << 
		file << ":" << line << std::endl;
	    std::cerr << "Error: " << error << std::endl;
	}
};
/********************************************************
 * internal_error -- Something went wrong that can't	*
 ********************************************************/
#define INTERNAL_ERROR(msg) throw(internal_error(__FILE__, __LINE__, msg))

class internal_error : public generic_error {
    public:

	explicit internal_error(
		const char *const i_file,
		const int i_line,
		const std::string &i_error):
	    generic_error(i_file, i_line, "Internal", i_error)
	{};
	// Copy constructor defaults
	// Assignment operator defaults
	// Destructor defaults
};

/********************************************************
 * user_error -- User did something wrong		*
 ********************************************************/
#define USER_ERROR(msg) throw(user_error(__FILE__, __LINE__, msg))

class user_error : public generic_error {
    public:

	explicit user_error(
		const char *const i_file,
		const int i_line,
		const std::string &i_error):
	    generic_error(i_file, i_line, "User", i_error)
	{};
	// Copy constructor defaults
	// Assignment operator defaults
	// Destructor defaults
};
/********************************************************
 * system_error -- System call returned unexpected	*
 * 	results.					*
 ********************************************************/
#define SYSTEM_ERROR(msg) throw(system_error(__FILE__, __LINE__, errno, msg))

class system_error : public generic_error {
    public:
	const int error_code;

	explicit system_error(
		const char *const i_file,
		const int i_line,
		const int i_error_code,
		const std::string &i_error):
	    generic_error(i_file, i_line, "System", i_error),
	    error_code(i_error_code)
	{};
	// Copy constructor defaults
	// Assignment operator defaults
	// Destructor defaults
	void print(void) {
	    generic_error::print();
	    std::cerr << "errno=" << error_code << " " <<
		sys_errlist[error_code] << std::endl;
	}
};

#endif /* __EXCEPTION_H__ */

