/********************************************************
 * filer -- An object to collect files and feed them	*
 * to the batcher.					*
 ********************************************************/
#include <unistd.h>
#include <algorithm>
#include <errno.h>
#include <dirent.h>

#include "exception.h"
#include "backup.h"
#include "log.h"

#include "filer.h"


/********************************************************
 * non_empty_dir -- Return true if the directory is	*
 * 	non-empty since we probably want to skip it.	*
 *********************************************************/
bool filer::non_empty_dir(
   std::string& dir_name,	// Name of the directory to check
   struct stat64& stat		// File mode
)
{
    if (!S_ISDIR(stat.st_mode)) {
	return (false);
    }
    // The directory
    DIR *dir_handle = opendir(dir_name.c_str());

    if (dir_handle == NULL) {
	log_msg(LOG_WARNING, "Directory " + dir_name + " protected.  Skipping.");
	return (true);
    }
    // The current director entry
    struct dirent *cur_entry;

    int counter = 0;	// Count the number of dir entries

    while (counter < 3) {
	cur_entry = readdir(dir_handle);
	if (cur_entry == NULL) {
	    closedir(dir_handle);
	    return (false);
	}
	++counter;
    }
    log_msg(LOG_DIR, 
	    "Warning: Skipping non-empty directory " + dir_name);

    closedir(dir_handle);
    return (true);
}
/********************************************************
 * filer::add_files -- Add the files to the list	*
 ********************************************************/
void filer::add_files(
    const string_set& files	// Files to add to the list we have
) {

    if (chdir(backup_root.c_str()) != 0) 
	SYSTEM_ERROR("chdir(" + backup_root + ") failed");

    // File we are working on now
    string_set::const_iterator cur_file;
    
    for (cur_file = files.begin();
	cur_file != files.end();
	++cur_file) {

	// The name of the file we are working on
	std::string file_name = *cur_file;

	if (file_name.at(0) == '/') {
	    file_name = file_name.substr(1);
	}

	// Did we do this one
	if (file_list.find(file_name) != file_list.end()) {
	    continue;
	}

	// Information about the file
	struct stat64 stat;

	if (lstat64(file_name.c_str(), &stat) != 0) {
	    log_msg(LOG_WARNING, "Warning " + file_name + " disappeared");
	    continue;
	}
	if (S_ISREG(stat.st_mode) || S_ISDIR(stat.st_mode)) {
	    if (access(file_name.c_str(), R_OK) != 0) {
		log_msg(LOG_WARNING, 
			"WARNING: Could not access " + file_name + " skipped");
		continue;
	    }
	}
	if (S_ISSOCK(stat.st_mode)) {
	    log_msg(LOG_WARNING, "WARNING: Skipping socket " + file_name);
	    continue;
	}

	if (S_ISFIFO(stat.st_mode)) {
	    log_msg(LOG_WARNING, "WARNING: Skipping fifo " + file_name);
	    continue;
	}
	if (non_empty_dir(file_name, stat)) {
	    continue;
	}
	file_list[file_name] = stat;
	id_list[id(stat)].insert(file_name);
    }
}

/********************************************************
 * all_names -- Given the name of a file, return all	*
 * the names it can be listed under.			*
 ********************************************************/
const string_set filer::all_names(
    const std::string& given_file	// File name of the current file
) const {
    // The pointer to the stat information
    file_to_stat::const_iterator stat = file_list.find(given_file);

    if (stat == file_list.end()) {
	INTERNAL_ERROR("Could not find " + 
		given_file + " in the file_list");
    }

    // The id for this file
    id file_id(stat->second);

    // All the files for this ID
    id_to_files::const_iterator cur_files = id_list.find(file_id);

    if (cur_files == id_list.end()) {
	INTERNAL_ERROR("Could not find id->files for file " + 
		given_file);
    }
    return (cur_files->second);
}

/********************************************************
 * filer::mark -- Mark the indicated files as done	*
 ********************************************************/
void filer::mark(
    const string_set& files	// List of files to mark as done
) {
    // Loop through each file
    for (string_set::const_iterator cur_file = files.begin();
	    cur_file != files.end();
	    ++cur_file) {
	mark(*cur_file);
    }
}
/********************************************************
 * filer::mark -- Mark the indicated files as done	*
 ********************************************************/
void filer::mark(
    const std::string& given_file	// List of files to mark as done
) {
    file_to_stat::iterator stat = file_list.find(given_file);

    if (stat == file_list.end()) {
	INTERNAL_ERROR("Could not find " + 
		given_file + " in the file_list");
    }
    file_list.erase(stat);
}

/********************************************************
 * filer::get_files -- Return the files in the filer	*
 ********************************************************/
const string_set filer::get_files(void) {
    string_set result;	// Result of our get

    // Loop through the array
    for (file_to_stat::const_iterator cur_file = file_list.begin();
	    cur_file != file_list.end();
	    ++cur_file) {
	result.insert(cur_file->first);
    }
    log_msg(LOG_PROGRESS, "Files left " + int_to_string(result.size()));
    return (result);
}
/********************************************************
 * done -- Return true if the file was done		*
 ********************************************************/
const bool filer::done(
    const std::string &file	// The file to check
)
{
    file_to_stat::iterator stat = file_list.find(file);
    return (stat == file_list.end());
}
