/********************************************************
 * meta_r -- Module to handle meta data on restore	*
 ********************************************************/
#include <fnmatch.h>
#include <cstdio>
#include <algorithm>

#include "util.h"
#include "log.h"
#include "colon.h"
#include "magic.h"
#include "exception.h"
#include "backup.h"
#include "meta_r.h"

/********************************************************
 * meta::restore -- Read in the meta data from 		*
 * a file.						*
 ********************************************************/
void meta_r::restore(
    const std::string& file_name
) {
    // The file containing the meta data
    std::ifstream in_file(file_name.c_str());

    if (in_file.bad()) 
	INTERNAL_ERROR("Could not open label file: " + file_name);

    // The line from the meta file
    std::string line;

    getline(in_file, line);
    if (line != HEADER_MAGIC)
	INTERNAL_ERROR("Bad magic line starting meta file");

    colon_line parser(in_file);	// Parse input line

    // Read the serial number
    parser.expect_line(META_SERIAL, 
	    "Error reading backup serial number from meta data",
	    __FILE__, __LINE__);
    log_msg(LOG_RESTORE, "Backup Serial number " + parser.get_value());
    
    parser.expect_line(META_ID, "Error reading meta file ID",
	    __FILE__, __LINE__);
    id = parser.get_value();
    log_msg(LOG_RESTORE, "Backup ID " + id);

    getline(in_file, line);	
    if (line != META_EOR)
	INTERNAL_ERROR("EOR Expected in meta restore");

    while (1) {
	getline(in_file, line);
	if (in_file.eof()) 
	    break;

	if (line != FILE_RECORD)
	    INTERNAL_ERROR("Expected file information, but got " + line);

	// Read the batch record from the file
	batch_info_r* batch_info = read_batch(in_file);
	io_info_r* io_info = read_io(in_file);
	info.insert(gen_info(batch_info, io_info));
    }
}
/********************************************************
 * meta::toc -- Generate a table of contents		*
 ********************************************************/
void meta_r::toc(
    std::ostream& out_file	// File to send the info to
) {
    create_file_data();

    // Previous file we printed
    std::string last_file("");

    // Loop through each file
    for (file_data_type::const_iterator cur_file = file_data.begin();
	  cur_file != file_data.end();
	  ++cur_file) {

	if (cur_file->first != last_file)
	    out_file << cur_file->first << std::endl;
	last_file = cur_file->first;
    }
}

/********************************************************
 * all_files -- Return a list of all the files in 	*
 * 	the backup.					*
 ********************************************************/
string_set meta_r::all_files(void)
{
    string_set result;

    create_file_data();

    // Previous file we printed
    std::string last_file("");

    // Loop through each file
    for (file_data_type::const_iterator cur_file = file_data.begin();
	  cur_file != file_data.end();
	  ++cur_file) {

	if (cur_file->first != last_file)
	    result.insert(cur_file->first);

	last_file = cur_file->first;
    }
    return (result);
}

/********************************************************
 * match -- Return a list of all the files that match	*
 * the glob strings in the "files" array.		*
 ********************************************************/
string_set meta_r::match(
    const string_set& files	// The list of files to check
) {
    string_set result;

    create_file_data();

    // Previous file we printed
    std::string last_file("");

    // Loop through each file
    for (file_data_type::const_iterator cur_file = file_data.begin();
	  cur_file != file_data.end();
	  ++cur_file) {

	if (cur_file->first != last_file) {
	    for (string_set::const_iterator cur_check = files.begin();
		 cur_check != files.end();
		 ++cur_check) {
		// Does this file match the wildcard?
		if (fnmatch(cur_check->c_str(), 
			    cur_file->first.c_str(),
			    FNM_PATHNAME) == 0) {
		    result.insert(cur_file->first);
		    break;
		}
	    }
	    last_file = cur_file->first;
	}
    }
    return (result);
}

/********************************************************
 * get_batch_set -- Return the set of batches that	*
 * 	match a given set of files			*
 ********************************************************/
meta_r::info_set meta_r::get_batch_set(
   const string_set& files		// Files we are restoring
) {
    info_set result;	// Restore of the work

    for (info_set::const_iterator cur_batch = info.begin();
	 cur_batch != info.end();
	 ++cur_batch) {

	// Check to see if a file in our file list
	// is contained in the current batch
	string_set::const_iterator search_result =
	    find_first_of(files.begin(), files.end(),
		    cur_batch->batch_info->files.begin(), 
		    cur_batch->batch_info->files.end());

	// Did this batch work?
	if (search_result != files.end()) {
	    log_msg(LOG_SET, "Adding batch " +
		    int_to_string(cur_batch->io_info->media_sequence) +
		    " file " + 
		    int_to_string(cur_batch->io_info->file_sequence) +
		    " batch " + 
		    int_to_string(cur_batch->batch_info->batch_number));
	    result.insert(*cur_batch);
	}
    }
    return (result);
}
/********************************************************
 * check_id -- Function object to check to see		*
 * if this batch is the current on we're looking for by *
 * id number.						*
 * 							*
 * See STL documentation / function object for 		*
 * why we are doing things this way.			*
 ********************************************************/
class check_id: 
    public std::binary_function<meta_r::gen_info, unsigned int, bool> 
{
    public:
	bool operator()(
	    meta_r::gen_info info,	// Information to check
	    unsigned int id		// Media to look for
	) const {
	    return (info.batch_info->batch_number == id);
	}
};

/********************************************************
 * add_ids -- Add the indicated ids to the stuff to 	*
 * 	restore.					*
 ********************************************************/
void meta_r::add_ids(
    std::set<int> ids, 		// The list of ids to add
    info_set& restore_set	// The set of things to add to
) {
    // Loop through each id to check
    for (std::set<int>::const_iterator cur_id = ids.begin();
	    cur_id != ids.end();
	    ++cur_id) {

	// Result of the search
	info_set::const_iterator search_result =
	    find_if(info.begin(), info.end(), 
		bind2nd(check_id(), *cur_id));

	if (search_result == info.end()) {
	    USER_ERROR("No such id " + int_to_string(*cur_id));
	}
	restore_set.insert(*search_result);
    }
}


