/********************************************************
 * proc -- Module to handle execution of background	
 * procedure.						
 *
 * This rather long module is designed to replace the
 * standard "system" function call.  But it does a 
 * lot more, that's why it is so long.
 *
 * All command output is logged using the log_msg
 * function call.   This includes stderr and stdout.
 * 
 * Secondly, there are several function calls available
 * to be used.  These are:
 *
 * do_system -- Execute a system command and wait for it 
 * 		to complete (just like system)
 *
 * bg_system -- Execute a system command in the background.
 * 		A callback function will be called when it is done.
 *
 * pipe_in -- Execute a command in the background and capture
 * 		the standard out in a pipe (which we can then read)
 *
 * pipe_out -- Execute a command in the background and connect the
 * 		standard in to a pipe which we can write
 *
 *
 * In addition to the these functions there is the do_poll which must
 * be called periodically in order to cause the I/O to happen.
 */
#include <sys/types.h>
#include <sys/poll.h>
#include <sys/wait.h>
#include <sys/resource.h>

#include <unistd.h>
#include <map>
#include <iostream>
#include <signal.h>
#include <fcntl.h>
#include <assert.h>
#include <stdio.h>

#include "exception.h"
#include "log.h"
#include "restore.h"
#include "exit.h"
#include "util.h"
#include "proc.h"

/*
 * Information about a forked off process
 */
struct fork_info {
    int fd[3];		// FD of the pipes to stdin, stdout, stderr
    pid_t pid;		// Pid of the new process
};
/*------------------------------------------------------*/
/* I/O section 						*/
/*------------------------------------------------------*/
// Function to call when the poll function detects input
typedef void poll_funct(void* cb_data);

/********************************************************
 * poll_info -- A pipe we are going to use for polling	*
 * and related data.  Registered pipes use this 	*
 * structure for polling.				*
 ********************************************************/
class poll_info {
    public:
	int fd;			// FD we are watching
	poll_funct* cb;		// What to call when input ready
	void* cb_data;		// Data to pass to cb when done
	struct pollfd *fd_ptr;	// Pointer to the fd entry in the poll array
    public:
	poll_info(void): fd(-1), cb_data(NULL), fd_ptr(NULL) {};
	poll_info(
		const int i_fd, 
		poll_funct* i_cb,
		void* i_cb_data
	    ):
		fd(i_fd), 
	        cb(i_cb), 
		cb_data(i_cb_data), 
		fd_ptr(NULL) 
	    {};
	// Copy constructor defaults
	// Assignment operator defaults
	// Destructor defaults
    public:
	void do_cb(void) {
	    assert(cb != NULL);
	    cb(cb_data);
	}
};

bool exit_poll;	// When set exit the polling loop

// A map that maps polling fds to polling data
typedef std::map<int, poll_info> poll_map_type;
static poll_map_type poll_map;

/********************************************************
 * register_poll_fd -- Register a fd which will cause	*
 * 	a function to be called when input is ready	*
 * 	on the fd.					*
 ********************************************************/
static void register_poll_fd(
    const int fd,	// FD we are watching
    poll_funct* cb,	// What to call when input ready
    void* cb_data	// Data to pass to cb when done
) {
    // The poll info we are adding
    poll_info new_info(fd, cb, cb_data);

    poll_map[fd]= new_info;
}
/********************************************************
 * un_register_poll_fd -- Unregister a fd 		*
 ********************************************************/
static void unregister_poll_fd(
    const int fd		// FD we are watching
) {
    // Pointer to the current entry in the map
    poll_map_type::iterator map_entry = poll_map.find(fd);

    if (map_entry == poll_map.end()) 
	INTERNAL_ERROR("Unregistering a fd that was not registered");

    poll_map.erase(map_entry);
}
/********************************************************
 * io_str::io_str -- Create an I/O stream and register	*
 * it wit the poll system.				*
 ********************************************************/
io_str::io_str(
       pid_t pid,		// Process ID of the process owning this
       const int i_fd, 		// FD of the file to read
       const char i_log_char	// Character to use for log_msg
): 
    fd(i_fd),
    log_char(i_log_char)
{
    char pid_str[100];	// String version of the pid message
    snprintf(pid_str, sizeof(pid_str), "[%d] ", pid);
    pid_msg = pid_str;

    register_poll_fd(i_fd, do_cb_static, static_cast<void*>(this));
}
/********************************************************
 * io_str::~io_str -- Destroy the I/O stream by 	*
 * removing this fd from the watch list.		*
 ********************************************************/
io_str::~io_str() {
    if (fd >= 0)
	unregister_poll_fd(fd);
}

/********************************************************
 * io_str::do_cb -- Call back that's called when I/O	*
 * is present on the given fd.  Record the data into 	*
 * a string, log file, and screen.			*
 ********************************************************/
void io_str::do_cb(void) 
{
    while (1) {
	// Buffer for the data from the input
	char buffer[100];

	memset(buffer, '\0', sizeof(buffer));

	// Size of the data we just read
	int size = read(fd, buffer, sizeof(buffer)-1);

	// If nothing left, exit
	if (size == 0)
	    break;

	if ((size < 0) && ((errno == EAGAIN) || (errno == EINTR)))
	    break;

	if (size < 0) 
	    SYSTEM_ERROR("Error reading pipe");

	// Add each character to the string.  
	for (char *cur_char = buffer; *cur_char != '\0'; ++cur_char) {
	    // If we reach the end of line, say so
	    if (*cur_char == '\n') {
		log_msg(log_char, pid_msg + str);
		str = "";
	    } else {
		str += *cur_char;
	    }
	}
    }
}

/********************************************************
 * do_poll -- Execute a poll statement repeatedly	*
 * 	until the global variable exit_poll is set.	*
 ********************************************************/
void do_poll(void)
{
    struct pollfd *fd_list = new pollfd[poll_map.size()];

    int poll_index = 0;	// Index into the current polling item

    // Loop trough each item in the list of stuff to poll and
    // assemble a list for the poll statement
    for (poll_map_type::iterator cur_poll = poll_map.begin();
	    cur_poll != poll_map.end();
	    ++cur_poll) {
	fd_list[poll_index].fd = cur_poll->second.fd;
	fd_list[poll_index].events = POLLIN;
	fd_list[poll_index].revents = 0;
	cur_poll->second.fd_ptr = &fd_list[poll_index];
	++poll_index;
    }


    // Poll and wait for I/O or an interrupt
    int status = poll(fd_list, poll_index, -1);

    if (status > 0) {
	// Loop through the register poll users looking for items to 
	// Execute a cb function on
	for (poll_map_type::iterator cur_poll = poll_map.begin();
		cur_poll != poll_map.end();
		++cur_poll) {
	    if (cur_poll->second.fd_ptr->revents != 0)
		cur_poll->second.do_cb();
	}
    }
    if (status >= 0) {
	delete []fd_list;
	return;
    }

    // We're allowed to be interrupted
    if (errno == EINTR) {
	delete []fd_list;
	return;
    }

    char  errno_str[100];	// Error number as a string
    snprintf(errno_str, sizeof(errno_str), 
	    "WARNING: Poll returned error: %d", errno);
    log_msg(LOG_WARNING, errno_str);
    std::cerr << errno_str << std::endl;
    delete []fd_list;
}
/*------------------------------------------------------*/
/* Command section follows 				*/
/*------------------------------------------------------*/

/********************************************************
 * system_info::system_info -- Init information on a 	*
 * 	pending "system" command.			*
 ********************************************************/
system_info::system_info(
    const pid_t i_pid,		// Process ID of the program
    const char i_log_char,	// Where we log things
    const int i_out_fd,		// FD for standard out
    const int i_err_fd,		// Error file descriptor
    sys_cb_funct *i_cb_funct	// Function to call when the item terminates
): 
	log_char(i_log_char),
	out_fd(i_out_fd),
	err_fd(i_err_fd),
	out_str(i_pid, i_out_fd, i_log_char), 
	err_str(i_pid, i_err_fd, i_log_char),
    	cb_funct(i_cb_funct),
	pid(i_pid),
	done(false),
	status(-1)
{}
system_info::system_info(
    const pid_t i_pid,		// Process ID of the program
    const char i_log_char,	// Where we log things
    const int i_out_fd,		// FD for standard out
    sys_cb_funct *i_cb_funct	// Function to call when the item terminates
): 
	log_char(i_log_char),
	out_fd(i_out_fd),
	err_fd(-1),
	out_str(i_pid, i_out_fd, i_log_char), 
    	cb_funct(i_cb_funct),
	pid(i_pid),
	done(false),
	status(-1)
{}

// Map pid->system_info
typedef std::map<pid_t, system_info*> proc_list_type;

// List of procedures we are waiting on
static proc_list_type proc_list;		

/********************************************************
 * child_handler -- Handle a SIGCHILD command.		*
 ********************************************************/
static void child_handler(int sig) 
{
    int status;	// Status of the child
    pid_t pid = waitpid(-1, &status, WNOHANG);	// Information on the child
    if (pid <= 0) {
	// If no child, return
	if (errno == ECHILD)
	    return;

	char err_str[100];	// Process as a string
	snprintf(err_str, sizeof(err_str), "%d", errno);
	log_msg(LOG_WARNING, 
		std::string("waitpid failed with error = ") + 
		    std::string(err_str));
	return;
    }

    proc_list_type::iterator cur_proc = proc_list.find(pid);
    if (cur_proc == proc_list.end()) {
	char proc_str[100];	// Process as a string
	snprintf(proc_str, sizeof(proc_str), "%d", pid);
	//TODO2: Remove the statement below when we get a handle on our
	// pipe process
	log_msg(LOG_PIPE, 
		std::string("NOTICE: Unknown process exited: ") + 
		int_to_string(pid) + 
		" with status " + int_to_string(status));
	log_msg(LOG_PIPE, "NOTICE: Possible pipe closed");
	return;
    }
    cur_proc->second->cb(pid, status);
}

/********************************************************
 * bg_system -- Execute a system command.		*
 * 	Starts the command and then calls a 		*
 * 	termination routine when done.			*
 ********************************************************/
system_info *bg_system(
   const std::string cmd,	// Command to execute
   const char log_ch,		// Character for logging
   sys_cb_funct *cb_funct	// Function to call when the item terminates
) {
    log_msg(log_ch, "Executing " + cmd);
    // What to do about a sigchild
    struct sigaction action_child;
    memset(&action_child, '\0', sizeof(action_child));
    action_child.sa_handler = child_handler;
    action_child.sa_flags = SA_NOCLDSTOP;

    if (sigaction(SIGCHLD, &action_child, NULL) != 0) 
	SYSTEM_ERROR("Could not set SIGCHLD signal");

    int stdout_pipe[2];		// Pipe from the child output
    int stderr_pipe[2];		// Pipe from the child input

    if (pipe(stdout_pipe) != 0) {
	SYSTEM_ERROR("Could not open stdout pipe");
    }
    if (pipe(stderr_pipe) != 0) {
	SYSTEM_ERROR("Could not open stderr pipe");
    }
    if (fcntl(stdout_pipe[0], F_SETFL, O_NONBLOCK) != 0) 
	SYSTEM_ERROR("Could not set non-blocking on child's stdout");

    if (fcntl(stderr_pipe[0], F_SETFL, O_NONBLOCK) != 0) 
	SYSTEM_ERROR("Could not set non-blocking on child's stderr");
    
    sigset_t child_set;	// Set of signals, with SIGCHILD set

    if (sigemptyset(&child_set) != 0)
	SYSTEM_ERROR("Could not create an empty signal set");

    if (sigaddset(&child_set, SIGCHLD) != 0) 
	SYSTEM_ERROR("Could not set the child bit");

    if (sigprocmask(SIG_BLOCK, &child_set, NULL) != 0) 
	SYSTEM_ERROR("Could not block SIGCHLD signal");

    pid_t pid = fork();	// Fork off a child process
    if (pid == 0) {
	// Setup the stdout, stderr for the command
	// Note: Stdin is left connected to the terminal.
	//    (Someone may want to read from the tty)
	//    If that's the case, they should lock the fd because
	//    other may read as well.

	//Note: Can't throw exceptions.  We're the child
	// Besides these will never fail -)
	if (dup2(stdout_pipe[1], 1) < 0) {
	    std::cerr << "dup2(stdout, 1) failed = " << errno << std::endl;
	    exit (EXIT_FATAL);
	}
	if (dup2(stderr_pipe[1], 2) < 0) {
	    std::cerr << "dup2(stderr, 2) failed = " << errno << std::endl;
	    exit (EXIT_FATAL);
	}

	struct rlimit rlimit;	// Resource limit
	if (getrlimit(RLIMIT_NOFILE, &rlimit) != 0) {
	    std::cerr << "Warning: getrlimit failed" << std::endl;
	} else {
	    // Close all legacy files
	    for (unsigned int fd = 3; fd < rlimit.rlim_cur; ++fd)
#ifdef VALGRIND
		if (fd != 99)
		    close(fd);
#else /* VALGRIND */
		close(fd);
#endif /* VALGRIND */
	}

	execl("/bin/sh", "sh", "-c", cmd.c_str(), NULL);
	std::cerr << "execl failed = " << errno << std::endl;
	exit (EXIT_FATAL);
    }
    // Register the new procedure
    system_info *info_ptr = 
	new system_info(pid, log_ch, 
		stdout_pipe[0], stderr_pipe[0], cb_funct);

    close(stdout_pipe[1]);
    close(stderr_pipe[1]);
    // Record the new pid in the list of things to worry about
    proc_list[pid] = info_ptr;

    // We're done adding a process, let the children finish 
    if (sigprocmask(SIG_UNBLOCK, &child_set, NULL) != 0) 
	SYSTEM_ERROR("Could not block SIGCHLD signal");
    return (info_ptr);
}
/********************************************************
 * log_exit -- Log the exit value of a system command	*
 ********************************************************/
static void log_exit(
    pid_t pid, 		// Pid of the process doing the command
    int status		// Status from wait containing exit code and other info
)
{
    char exit_msg[100];	// String version of the message we are sending

    snprintf(exit_msg, sizeof(exit_msg),
	    "Pid: %d Exited with status %x",
	    pid, status);
    log_msg(LOG_INFO, exit_msg);
}

/********************************************************
 * do_system -- Execute a system command and return	*
 * the result.  All output is logged.			*
 ********************************************************/
int do_system(
   const std::string cmd,	// Command to execute
   const char log_ch		// Character for logging
) 
{
    log_msg(LOG_CMD, "Cmd: " + cmd);

    // Get my process id
    system_info* info_ptr = bg_system(cmd, log_ch, log_exit);

    // Loop until the process we started
    // exits
    while (1) {
	if (info_ptr->done)
	    break;
	do_poll();
    }

    // Status from the command
    int status = info_ptr->status;

    // Pointer to the proc information for this process
    proc_list_type::iterator my_proc = proc_list.find(info_ptr->pid);
    assert(my_proc != proc_list.end());

    delete info_ptr;
    info_ptr = NULL;

    proc_list.erase(my_proc);
    return (status);
}
/********************************************************
 * fork_off -- Fork off a child process			*
 * 							*
 * Returns information about the child process		*
 ********************************************************/
static fork_info fork_off(
    const std::string &cmd,	// Command to execute
    const char log_ch		// Log character
) 
{
    fork_info result;		// Information about the fork

    int stdin_pipe[2];		// Pipe to the child stdin
    int stdout_pipe[2];		// Pipe from the child output
    int stderr_pipe[2];		// Pipe from the child input

    if (pipe(stdin_pipe) != 0) {
	SYSTEM_ERROR("Could not open stdin pipe");
    }

    if (pipe(stdout_pipe) != 0) {
	SYSTEM_ERROR("Could not open stdout pipe");
    }
    if (pipe(stderr_pipe) != 0) {
	SYSTEM_ERROR("Could not open stderr pipe");
    }
    if (fcntl(stdout_pipe[0], F_SETFL, O_NONBLOCK) != 0) 
	SYSTEM_ERROR("Could not set non-blocking on child's stdout");

    if (fcntl(stderr_pipe[0], F_SETFL, O_NONBLOCK) != 0) 
	SYSTEM_ERROR("Could not set non-blocking on child's stderr");

    pid_t pid = fork();	// Fork off a child process
    if (pid == 0) {
	// Setup the stdout, stderr for the command
	// Note: Stdin is left connected to the terminal.
	//    (Someone may want to read from the tty)
	//    If that's the case, they should lock the fd because
	//    other may read as well.

	//Note: Can't throw exceptions.  We're the child
	// Besides these will never fail -)
	if (dup2(stdout_pipe[0], 0) < 0) {
	    std::cerr << "dup2(stdin, 0) failed = " << errno << std::endl;
	    exit (EXIT_FATAL);
	}
	if (dup2(stdout_pipe[1], 1) < 0) {
	    std::cerr << "dup2(stdout, 1) failed = " << errno << std::endl;
	    exit (EXIT_FATAL);
	}
	if (dup2(stderr_pipe[1], 2) < 0) {
	    // Where would this error message go if we got there?
	    std::cerr << "dup2(stderr, 2) failed = " << errno << std::endl;
	    exit (EXIT_FATAL);
	}

	/*
	 * Close all the other files in the system
	 */
	struct rlimit rlimit;	// Resource limit
	if (getrlimit(RLIMIT_NOFILE, &rlimit) != 0) {
	    std::cerr << "Warning: getrlimit failed" << std::endl;
	} else {
	    // Close all legacy files
	    for (unsigned int fd = 3; fd < rlimit.rlim_cur; ++fd)
#ifdef VALGRIND
		if (fd != 99)
		    close(fd);
#else /* VALGRIND */
		close(fd);
#endif /* VALGRIND */
	}

	execl("/bin/sh", "sh", "-c", cmd.c_str(), NULL);
	std::cerr << "execl failed = " << errno << std::endl;
	exit (EXIT_FATAL);
    }
    result.fd[0] = stdin_pipe[1];
    result.fd[1] = stdout_pipe[0];
    result.fd[2] = stdout_pipe[0];
    result.pid = pid;

    close(stdout_pipe[0]);
    close(stdout_pipe[1]);
    close(stderr_pipe[1]);
    return (result);
}

/********************************************************
 * child_off -- turn off the SIGCHLD signal until	*
 * 	we get finished with forking things, then we	*
 * 	should turn it on with child_on.		*
 ********************************************************/
static void child_off(void)
{
    // What to do about a sigchild
    struct sigaction action_child;
    memset(&action_child, '\0', sizeof(action_child));
    action_child.sa_handler = child_handler;
    action_child.sa_flags = SA_NOCLDSTOP;

    if (sigaction(SIGCHLD, &action_child, NULL) != 0) 
	SYSTEM_ERROR("Could not set SIGCHLD signal");

    sigset_t child_set;	// Set of signals, with SIGCHILD set

    if (sigemptyset(&child_set) != 0)
	SYSTEM_ERROR("Could not create an empty signal set");

    if (sigaddset(&child_set, SIGCHLD) != 0) 
	SYSTEM_ERROR("Could not set the child bit");

    if (sigprocmask(SIG_BLOCK, &child_set, NULL) != 0) 
	SYSTEM_ERROR("Could not block SIGCHLD signal");
}

/********************************************************
 * child_on -- Turn on the SIGCHLD interrupt		*
 ********************************************************/
static void child_on(void)
{
    sigset_t child_set;	// Set of signals, with SIGCHILD set

    if (sigemptyset(&child_set) != 0)
	SYSTEM_ERROR("Could not create an empty signal set");

    if (sigaddset(&child_set, SIGCHLD) != 0) 
	SYSTEM_ERROR("Could not set the child bit");

    // We're done adding a process, let the children finish 
    if (sigprocmask(SIG_UNBLOCK, &child_set, NULL) != 0) 
	SYSTEM_ERROR("Could not block SIGCHLD signal");
}
/********************************************************
 * pipe_in -- Open a pipe from which we read		*
 * 	the output of the child.
 * 							*
 * Returns: the fd of the pipe 				*
 *********************************************************/
int pipe_in(
    const std::string &cmd,	// Command to execute
    const char log_ch,		// Log character
    sys_cb_funct *cb_funct	// Function to call when the item terminates
) {
    child_off();

    fork_info info = fork_off(cmd, log_ch);

    // Pointer to the info record for this process
    system_info *info_ptr = 
	new system_info(info.pid, log_ch, info.fd[2], cb_funct);

    close(info.fd[0]);	// Close the pipe going out
    proc_list[info.pid] = info_ptr;

    child_on();
    return (info.fd[1]);
}


