/********************************************************
 * slice_batch_b -- Module to backup files as a 	*
 * 	sliced tar file.				*
 ********************************************************/
#include <unistd.h>
#include <sys/fcntl.h>

#include "magic.h"
#include "slice_batch_b.h"
#include "util.h"
#include "backup.h"
#include "tmp_file.h"
#include "crc32.h"
#include "backup_info.h"
#include "log.h"

/********************************************************
 * slice_batch_info -- Information on a tar batch	*
 ********************************************************/
class slice_batch_info_b: public batch_info_b {
    public:
	unsigned int slice;	// Slice number
	bool last;		// True if this is the last slice

    public:
	explicit slice_batch_info_b(
	     const string_set& i_files,
	     const std::string& i_batch_file) :
	    batch_info_b(BATCH_TYPE_SLICE, i_files, i_batch_file) {}
	// Destructor defaults
	// Copy constructor defaults
	// Assignment operator defaults
	void write(std::ostream &out_file) const;
};

/********************************************************
 * slice_batch::slice_batch -- Create a slice batcher	*
 ********************************************************/
slice_batch_b::slice_batch_b(
    const std::string& b_name,	// The name of the of the batcher
    const config& config	// Configuration information
) : batch_b(b_name, config) 
{
    int block_size = 20 * 512;	// Tape block size (default = 20b)

    if (block_restriction != 0)
	block_size = block_restriction;

    slice_size = convert_size(
	    config.get_option(SECTION_BATCH, b_name, BATCH_SIZE));

    // to make sure we have a even number of tape blocks
    slice_size = static_cast<int>(slice_size / block_size) * block_size;

    buffer_limit = 62*1024;	// The number of bytes to transfer at one time
    if (config.defined(SECTION_BATCH, b_name, BATCH_BUFFER_SIZE)) {
	buffer_limit = convert_size(
		config.get_option(SECTION_BATCH, b_name, BATCH_BUFFER_SIZE));
    }
    sequence = 0;		// Batch file sequence number
}

/********************************************************
 * slice_batch::add_files -- Add a set of files to the	*
 * batch.  Actually this adds a single file and all 	*
 * it's links.						*
 ********************************************************/
batch_b::status slice_batch_b::add_files(
    filer &the_filer	// The filer which contains the files to use
)
{
    if (! names_needed) {
	if (done) {
	    return (BATCH_NONE);
	}
    }

    // Get a list of files to add
    string_set files_to_add = the_filer.get_files();

    last = false;
    
    // Loop through each possible file
    for (string_set::const_iterator cur_file = files_to_add.begin();
	    cur_file != files_to_add.end();
	    ++cur_file) {
	struct stat64 the_stat;	// The stat information for the file

	if (lstat64(cur_file->c_str(), &the_stat) != 0) {
	    log_msg(LOG_WARNING, 
		    "Warning: File went away:" + *cur_file);
	    continue;
	}
	if (!S_ISREG(the_stat.st_mode)) {
	    continue;	// We only work on files
	}
	current_files = the_filer.all_names(*cur_file);
	for (string_set::const_iterator add_file = current_files.begin();
	     add_file != current_files.end();
	     ++add_file)
	{
	    log_msg(LOG_BATCH, "Adding file to batch " + *add_file);
	}
	break;	// We're done with finding the file to backup
    }
    if (current_files.size() == 0) {
	return (BATCH_NONE);
    }

    std::string cmd;	// Command to go the pipe
    if (names_needed) {
	// The name of the file containing the names
	name_file_name = write_name_file("slice.names", current_files);

	cmd = "cd " + backup_root + ";name_file=" + name_file_name +
	    ";" + backup_cmd;
    } else {
	cmd = "cd " + backup_root + ";" + backup_cmd;
    }

    batch_number = next_batch_number;
    ++next_batch_number;
    pipe_file = popen(cmd.c_str(), "r");
    if (pipe_file == NULL) {
	SYSTEM_ERROR("Could not open pipe to " + cmd);
    }
    sequence = 0;
    the_filer.mark(current_files);
    return (BATCH_FULL);
}

/********************************************************
 * slice_batch::create_batch -- Create the batch file	*
 * from slicing the input tar stream.			*
 ********************************************************/
batch_info_b* slice_batch_b::create_batch(void) 
{
    if (current_files.size() == 0) 
	return (NULL);
    
    // The name of the file containing the slice
    std::string batch_file = tmp_name("slice.file");

    // The file to write the slice to
    int slice_file = open(batch_file.c_str(), O_WRONLY|O_CREAT, 0600);
    if (slice_file <= 0)
	SYSTEM_ERROR("Could not create " + batch_file);

    // The size of the data copied (adjusted for the stuff we do automatically)
    size_t copy_size = sizeof(struct file_head) + sizeof(struct file_tail);

    // The buffer for copying the files
    char *buffer = new char[buffer_limit];

    struct file_head head = {0};
    head.magic = HEAD_MAGIC;
    head.struct_size = sizeof(head);
    head.backup_number = backup_serial();
    if (write(slice_file, &head, sizeof(head)) != sizeof(head))
	SYSTEM_ERROR("Could not write slice output file");
    
    while (copy_size < slice_size) {
	// The number of bytes left for the copy
	int bytes_left = slice_size - copy_size;

	if (bytes_left > static_cast<long int>(buffer_limit))
	    bytes_left = static_cast<long int>(buffer_limit);

	// The size of the actual read
	int size = fread(buffer, 1, bytes_left, pipe_file);
	if (size <= 0) {
	    last = 1;
	    break;
	}
	if (write(slice_file, buffer, size) != size)
	    SYSTEM_ERROR("Could not write slice output file");
	copy_size += size;
	head.size += size;
	head.crc = crc32(head.crc, buffer, size);
    }
    delete [] buffer;

    struct file_tail tail = {0};
    tail.magic = TAIL_MAGIC;
    tail.struct_size = sizeof(tail);
    if (write(slice_file, &tail, sizeof(tail)) != sizeof(tail))
	SYSTEM_ERROR("Could not write slice output file");

    if (lseek(slice_file, 0, SEEK_SET) != 0L)
	SYSTEM_ERROR("Could not seek on slice output file");

    head.slice_number = sequence;
    head.last = last;
    if (write(slice_file, &head, sizeof(head)) != sizeof(head))
	SYSTEM_ERROR("Could not write slice output file");

    close(slice_file);

    if (last) {
	pclose(pipe_file);

	if (names_needed) {
	    if (! keep_option) {
		if (unlink(name_file_name.c_str()) != 0)
		    SYSTEM_ERROR("Unable to remove " + name_file_name);
	    }
	} else {
	    done = true;
	}
    }
    // The information on this batch file
    // Decallocated in the IO module
    slice_batch_info_b *info = 
	new slice_batch_info_b(current_files, batch_file);

    info->slice = sequence;
    info->last = last;
    info->batch_number = batch_number;
    info->name = name;

    info->backup_cmd = backup_cmd;
    info->verify_cmd = verify_cmd;
    info->restore_cmd = restore_cmd;
    info->names_needed = names_needed;

    if (last)
	current_files.clear();
    ++sequence;
    return (info);
}

/********************************************************
 * slice_batch_info::write -- Write out the information	*
 * to a file.						*
 ********************************************************/
void slice_batch_info_b::write(
   std::ostream& out_file	// File to write the stuff to
) const {
    write_base(out_file);
    out_file << META_BATCH_SLICE << ": " << slice << std::endl;
    out_file << META_BATCH_LAST << ": " << last << std::endl;
    out_file << META_EOR << std::endl;
}
