/********************************************************
 * util -- General purpose utilities 			*
 ********************************************************/
#include <sys/types.h>
#include <ctype.h>
#include <cstring>
#include <cstdio>
#include <fstream>

#include "backup.h"
#include "util.h"
#include "exception.h"
#include "tmp_file.h"
/********************************************************
 * convert_size -- Convert the size of something 	*
 * into bytes.						*
 *							*
 * Knows about the suffixes MB, K, and b for Megabyte, 	*
 * Kbyte, and block (512 bytes)				*
 ********************************************************/
off_t convert_size(
   const std::string &size	// Size to convert
)
{
    unsigned int split;	// Where to split the digits and suffix

    for (split = 0; split < size.length(); ++split) {
	if (! isdigit(size.at(split)))
	    break;
    }
    if (split == 0) 
	USER_ERROR("Size expected.  Got " + size);

    // The digits parts of the string
    std::string digits = size.substr(0, split);

    // The suffix of the string
    std::string suffix = size.substr(split);

    // The size in bytes
    off_t size_bytes = atol(digits.c_str());

    if (suffix.c_str() == "") {
	return (size_bytes);
    }
    if (strcasecmp(suffix.c_str(), "b") == 0) {
	return (size_bytes * 512);
    }
    if (strcasecmp(suffix.c_str(), "k") == 0) {
	return (size_bytes * 1024);
    }
    if (strcasecmp(suffix.c_str(), "mb") == 0) {
	return (size_bytes * 1024 * 1024);
    }
    if (strcasecmp(suffix.c_str(), "m") == 0) {
	return (size_bytes * 1024 * 1024);
    }
    USER_ERROR("Unknown prefix for size: " + size);
}
/********************************************************
 * encode_string -- Encode funny characters in a string	*
 * 	so it can be printed to a file without 		*
 * 	worrying about them.				*
 * 							*
 * Encoding consists of replacing specials with =xx	*
 * where xx are hex digits.				*
 ********************************************************/
std::string encode_string(
    const std::string& input	// String to encode
) {
    std::string output;		// Resulting string

    // Loop through the input looking for funnies
    for (std::string::const_iterator cur_char = input.begin();
	    cur_char != input.end();
	    ++cur_char) {
	if ((*cur_char < ' ') || (*cur_char == '=')) {
	    char code[8];

	    sprintf(code, "=%02X", (*cur_char));
	    output += code;
	} else {
	    output += (*cur_char);
	}
    }
    return (output);
}
/********************************************************
 * decode_string -- Decode funny characters in a string	*
 * 							*
 * Encoding consists of replacing specials with =xx	*
 * where xx are hex digits.				*
 ********************************************************/
std::string decode_string(
    const std::string& input	// String to decode
) {
    std::string output;		// Resulting string

    // Loop through the input looking for funnies
    for (std::string::const_iterator cur_char = input.begin();
	    cur_char != input.end();
	    ++cur_char) {

	if (*cur_char == '=') {
	    char code[3];	// Character as a 2 character hex cod
	    int i_char;		// Character as an integer

	    code[0] = *(cur_char+1);
	    code[1] = *(cur_char+2);
	    code[2] = '\0';
	    if (sscanf(code, "%02X", &i_char) != 1) 
		INTERNAL_ERROR("sscanf returned funny result");
	    output += static_cast<char>(i_char);
	    cur_char += 2;
	} else {
	    output += (*cur_char);
	}
    }
    return (output);
}

/********************************************************
 * write_name_file -- Write out a tar name file		*
 *  Entries end with \0 for use with the --null		*
 *  							*
 * Returns: 						*
 * 	The name of the temporary file.			*
 ********************************************************/
std::string write_name_file(
   const char *const basename,		// The base name of the temp file
   const string_set file_list		// The list of file names
) {
    // The name of the file to hold the names
    std::string name_file_name = tmp_name(basename);

    // The file to put the names into
    std::ofstream name_file(name_file_name.c_str());

    // Loop through each file and write it out
    for (string_set::const_iterator cur_file = file_list.begin();
	    cur_file != file_list.end();
	    ++cur_file)
    {
	name_file << *cur_file << '\0';
    }
    name_file.close();
    return (name_file_name);
}

/********************************************************
 * int_to_string -- Turn an integer into a string	*
 ********************************************************/
std::string int_to_string(
    const int value	// Value to convert
) {
    char str[100];	// String version of the value
    snprintf(str, sizeof(str), "%d", value);
    return (std::string(str));
}
